import argparse
import importlib
import inspect
import json
import logging
import os.path
import sys
from copy import deepcopy
from dataclasses import dataclass, field, fields, replace
from typing import (
    Callable,
    Optional,
    TextIO,
)

from databricks.bundles.core import Bundle, Diagnostics, Location, Resources
from databricks.bundles.core._resource_mutator import ResourceMutator
from databricks.bundles.core._resource_type import _ResourceType
from databricks.bundles.core._transform import _transform

__all__ = []


@dataclass
class _Args:
    phase: str
    input: str
    output: str
    diagnostics: str
    locations: Optional[str]  # FIXME should be required
    unknown_args: list[str]


@dataclass
class _Conf:
    resources: list[str] = field(default_factory=list)
    mutators: list[str] = field(default_factory=list)
    venv_path: Optional[str] = None

    @classmethod
    def from_dict(cls, d: dict) -> "_Conf":
        known_keys = [f.name for f in fields(cls)]
        unknown_keys = d.keys() - known_keys

        if unknown_keys:
            logging.warning(f"Unknown configuration keys: {unknown_keys}")

        return _transform(cls, {k: v for k, v in d.items() if k in known_keys})


def _parse_args(args: list[str]) -> _Args:
    parser = argparse.ArgumentParser()
    parser.add_argument("--phase", default=None)
    parser.add_argument("--input", default=None)
    parser.add_argument("--output", default=None)
    parser.add_argument("--diagnostics", default=None)
    parser.add_argument("--locations", default=None)

    parsed, unknown_args = parser.parse_known_args(args)

    if not parsed.phase:
        raise ValueError("Missing required argument --phase")

    if not parsed.input:
        raise ValueError("Missing required argument --input")

    if not parsed.output:
        raise ValueError("Missing required argument --output")

    if not parsed.diagnostics:
        raise ValueError("Missing required argument --diagnostics")

    return _Args(
        phase=parsed.phase,
        input=parsed.input,
        output=parsed.output,
        diagnostics=parsed.diagnostics,
        locations=parsed.locations,
        unknown_args=unknown_args,
    )


def _load_resources_from_input(input: dict) -> tuple[Resources, Diagnostics]:
    resources = Resources()
    diagnostics = Diagnostics()

    input_resources = input.get("resources", {})

    for tpe in _ResourceType.all():
        input_resources_by_tpe = input_resources.get(tpe.plural_name, {})

        for resource_name, resource_dict in input_resources_by_tpe.items():
            try:
                resource = _transform(tpe.resource_type, resource_dict)

                resources.add_resource(resource_name=resource_name, resource=resource)
            except Exception as exc:
                diagnostics = diagnostics.extend(
                    Diagnostics.from_exception(
                        exc=exc,
                        summary=f"Error while loading {tpe.singular_name}",
                        path=("resources", tpe.plural_name, resource_name),
                    )
                )

    return resources, diagnostics


def _apply_mutators(
    bundle: Bundle,
    resources: Resources,
    mutator_functions: list[ResourceMutator],
) -> tuple[Resources, Diagnostics]:
    diagnostics = Diagnostics()

    for tpe in _ResourceType.all():
        resources, diagnostics = diagnostics.extend_tuple(
            _apply_mutators_for_type(bundle, resources, tpe, mutator_functions)
        )

    return resources, diagnostics


def _apply_mutators_for_type(
    bundle: Bundle,
    resources: Resources,
    tpe: _ResourceType,
    mutator_functions: list[ResourceMutator],
) -> tuple[Resources, Diagnostics]:
    resources_dict = getattr(resources, tpe.plural_name)

    for resource_name, resource in resources_dict.items():
        for mutator in mutator_functions:
            if mutator.resource_type != tpe.resource_type:
                continue

            location = Location.from_callable(mutator.function)

            try:
                if _get_num_args(mutator.function) == 1:
                    new_resource = mutator.function(resource)
                else:
                    # defensive copy so that one function doesn't affect another
                    new_resource = mutator.function(deepcopy(bundle), resource)

                # mutating resource in-place works, but we can't tell when it happens,
                # so we only update location if new instance is returned

                if new_resource is not resource:
                    if location:
                        resources.add_location(
                            ("resources", tpe.plural_name, resource_name), location
                        )
                    resources_dict[resource_name] = new_resource
                    resource = new_resource
            except Exception as exc:
                mutator_name = mutator.function.__name__

                return resources, Diagnostics.from_exception(
                    exc=exc,
                    summary=f"Failed to apply '{mutator_name}' mutator",
                    location=location,
                    path=("resources", tpe.plural_name, resource_name),
                )

    return resources, Diagnostics()


def _read_conf(input: dict) -> tuple[_Conf, Diagnostics]:
    experimental = input.get("experimental", {})

    if experimental.get("pydabs", {}) != {}:
        return (
            _Conf(),
            Diagnostics.create_error(
                "'experimental/pydabs' is not supported by 'databricks-bundles', use 'python' instead",
                detail="",
                location=None,
                path=("experimental", "pydabs"),
            ),
        )

    experimental_conf_dict = experimental.get("python", {})
    experimental_conf = _transform(_Conf, experimental_conf_dict)

    conf_dict = input.get("python", {})
    conf = _transform(_Conf, conf_dict)

    has_conf = conf != _Conf()
    has_experimental_conf = experimental_conf != _Conf()

    if has_conf and not has_experimental_conf:
        return conf, Diagnostics()
    elif not has_conf and has_experimental_conf:
        # do not generate warning in Python code, if CLI supports non-experimental 'python',
        # it should generate a warning
        return experimental_conf, Diagnostics()
    elif has_conf and has_experimental_conf:
        # for backward-compatibility, CLI can copy contents of 'python' into 'experimental/python'
        # if configs are equal, it isn't a problem
        if conf != experimental_conf:
            return (
                _Conf(),
                Diagnostics.create_error(
                    "Both 'python' and 'experimental/python' sections are present, use 'python' section only",
                    detail="",
                    location=None,
                    path=(
                        "experimental",
                        "python",
                    ),
                ),
            )
        else:
            return conf, Diagnostics()
    else:
        return _Conf(), Diagnostics()


def python_mutator(
    args: _Args,
) -> tuple[dict, dict[tuple[str, ...], Location], Diagnostics]:
    input = json.load(open(args.input, encoding="utf-8"))
    diagnostics = Diagnostics()

    conf, diagnostics = diagnostics.extend_tuple(_read_conf(input))
    if diagnostics.has_error():
        return input, {}, diagnostics

    bundle = _parse_bundle_info(input)

    if args.phase == "load_resources":
        resource_functions, diagnostics = _load_functions(conf.resources)
        if diagnostics.has_error():
            return input, {}, diagnostics

        resources, diagnostics = diagnostics.extend_tuple(
            _load_resources(bundle, resource_functions)
        )
        if diagnostics.has_error():
            return input, {}, diagnostics

        new_bundle = _append_resources(input, resources)
        locations = _relativize_locations(resources._locations)

        return new_bundle, locations, diagnostics.extend(resources.diagnostics)
    elif args.phase == "apply_mutators":
        mutator_functions, diagnostics = _load_resource_mutators(conf.mutators)
        if diagnostics.has_error():
            return input, {}, diagnostics

        resources, diagnostics = _load_resources_from_input(input)
        if diagnostics.has_error():
            return input, {}, diagnostics

        resources, diagnostics = diagnostics.extend_tuple(
            _apply_mutators(bundle, resources, mutator_functions)
        )
        if diagnostics.has_error():
            return input, {}, diagnostics

        new_bundle = _append_resources(input, resources)
        locations = _relativize_locations(resources._locations)

        return new_bundle, locations, diagnostics.extend(resources.diagnostics)
    else:
        return input, {}, Diagnostics.create_error(f"Unknown phase: {args.phase}")


def _parse_bundle_info(input: dict) -> Bundle:
    bundle = input.get("bundle", {})
    variables = {k: v.get("value") for k, v in input.get("variables", {}).items()}

    return Bundle(
        target=bundle["target"],
        variables=variables,
    )


def _append_resources(bundle: dict, resources: Resources) -> dict:
    """
    Append resources to input without modifying resources that are already present.
    """

    new_bundle = bundle.copy()

    for tpe in _ResourceType.all():
        resources_dict = getattr(resources, tpe.plural_name)

        if resources_dict:
            new_bundle["resources"] = new_bundle.get("resources", {})
            new_bundle["resources"][tpe.plural_name] = new_bundle["resources"].get(
                tpe.plural_name, {}
            )

            for resource_name, resource in resources_dict.items():
                new_bundle["resources"][tpe.plural_name][resource_name] = (
                    resource.as_dict()
                )

    return new_bundle


def _get_num_args(func: Callable) -> int:
    return len(inspect.signature(func).parameters)


def _load_resources(
    bundle: Bundle,
    functions: list[Callable],
) -> tuple[Resources, Diagnostics]:
    diagnostics = Diagnostics()
    resources = Resources()

    for function in functions:
        try:
            function_resources, diagnostics = diagnostics.extend_tuple(
                _load_resources_from_function(bundle, function)
            )

            resources.add_resources(function_resources)
        except Exception as exc:
            diagnostics = diagnostics.extend(
                Diagnostics.from_exception(
                    exc=exc,
                    summary="Failed to load resources",
                    location=Location.from_callable(function),
                )
            )

    return resources, diagnostics


def _load_functions(names: list[str]) -> tuple[list[Callable], Diagnostics]:
    diagnostics = Diagnostics()
    functions = []

    for name in names:
        try:
            function, diagnostics = diagnostics.extend_tuple(_load_function(name))

            if function:
                functions.append(function)
        except Exception as exc:
            diagnostics = diagnostics.extend(
                Diagnostics.from_exception(
                    exc=exc,
                    summary=f"Failed to load function '{name}'",
                )
            )

    return functions, diagnostics


def _load_resource_mutators(
    names: list[str],
) -> tuple[list[ResourceMutator], Diagnostics]:
    diagnostics = Diagnostics()
    functions = []

    for name in names:
        try:
            function, diagnostics = diagnostics.extend_tuple(
                _load_resource_mutator(name)
            )

            if function:
                functions.append(function)
        except Exception as exc:
            diagnostics = diagnostics.extend(
                Diagnostics.from_exception(
                    exc=exc,
                    summary=f"Failed to load mutator '{name}'",
                )
            )

    return functions, diagnostics


def _load_object(qualified_name: str) -> tuple[Optional[Callable], Diagnostics]:
    diagnostics = Diagnostics()
    [module_name, name] = qualified_name.split(":")

    common_error = qualified_name == "resources:load_resources"

    try:
        module = importlib.import_module(module_name)
    except Exception as exc:
        if isinstance(exc, ModuleNotFoundError) and exc.name == "resources":
            if common_error:
                return None, _explain_common_import_error(exc)

        return None, Diagnostics.from_exception(
            exc=exc,
            summary=f"Failed to import module '{module_name}'",
        )

    try:
        return getattr(module, name), diagnostics
    except AttributeError as exc:
        if common_error:
            return None, _explain_common_import_error(exc)

        return None, Diagnostics.from_exception(
            exc=exc,
            summary=f"Name '{name}' not found in module '{module_name}'",
            location=Location(file=module.__file__) if module.__file__ else None,
        )
    except Exception as exc:
        return None, Diagnostics.from_exception(
            exc=exc,
            summary=f"Failed to load function '{name}' from module '{module_name}'",
        )


def _explain_common_import_error(exc: Exception) -> Diagnostics:
    # a common case when default name of the module is not found
    # we can give a hint to the user how to fix it
    explanation = """Make sure to create a new Python file at resources/__init__.py with contents:

from databricks.bundles.core import load_resources_from_current_package_module


def load_resources():
    return load_resources_from_current_package_module()
"""
    return Diagnostics.from_exception(
        exc=exc,
        summary="Can't find function 'load_resources' in module 'resources'",
        explanation=explanation,
    )


def _load_function(qualified_name: str) -> tuple[Optional[Callable], Diagnostics]:
    [instance, diagnostics] = _load_object(qualified_name)
    [module_name, name] = qualified_name.split(":")

    if diagnostics.has_error():
        return None, diagnostics

    if instance and not isinstance(instance, Callable):
        return None, Diagnostics.create_error(
            f"Function '{name}' in module '{module_name}' is not callable",
        )

    return instance, diagnostics


def _load_resource_mutator(
    qualified_name: str,
) -> tuple[Optional[ResourceMutator], Diagnostics]:
    [instance, diagnostics] = _load_object(qualified_name)
    [module_name, name] = qualified_name.split(":")

    if diagnostics.has_error():
        return None, diagnostics

    if instance and not isinstance(instance, ResourceMutator):
        return None, Diagnostics.create_error(
            f"'{name}' in module '{module_name}' is not instance of ResourceMutator, did you decorate it with @<resource_type>_mutator?",
        )

    return instance, diagnostics


def _load_resources_from_function(
    bundle: Bundle,
    function: Callable,
) -> tuple[Resources, Diagnostics]:
    if _get_num_args(function) == 0:
        resources = function()
    else:
        # defensive copy so that one function doesn't affect another
        resources = function(deepcopy(bundle))

    assert isinstance(resources, Resources)

    return resources, Diagnostics()


def main(argv: list[str]) -> int:
    args = _parse_args(argv[1:])

    if args.unknown_args:
        logging.warning(f"Unknown arguments: {args.unknown_args}")

    logging.basicConfig(level=logging.DEBUG, stream=sys.stderr)
    new_bundle, locations, diagnostics = python_mutator(args)

    with open(args.diagnostics, "w", encoding="utf-8") as f:
        _write_diagnostics(f, diagnostics)

    if locations_path := args.locations:
        with open(locations_path, "w", encoding="utf-8") as f:
            _write_locations(f, locations)

    with open(args.output, "w", encoding="utf-8") as f:
        _write_output(f, new_bundle)

    return 1 if diagnostics.has_error() else 0


def _write_diagnostics(f: TextIO, diagnostics: Diagnostics) -> None:
    for diagnostic in diagnostics.items:
        obj = diagnostic.as_dict()

        if obj.get("path"):
            obj["path"] = ".".join(obj["path"])

        json.dump(obj, f, ensure_ascii=False)
        f.write("\n")


def _write_output(f: TextIO, bundle: dict) -> None:
    json.dump(bundle, f, ensure_ascii=False)


def _relativize_locations(
    locations: dict[tuple[str, ...], Location],
) -> dict[tuple[str, ...], Location]:
    return {
        path: _relativize_location(location) for path, location in locations.items()
    }


def _relativize_location(location: Location) -> Location:
    return replace(location, file=_relativize_path(location.file))


def _relativize_path(path: str) -> str:
    """
    Attempt to relativize an absolute path to the current working directory.

    If the path is not absolute or cannot be relativized, return it as is.
    Used to relativize paths in locations to show shorter paths in diagnostics.
    """

    if not os.path.isabs(path):
        return path

    cwd = os.getcwd()

    try:
        common = os.path.commonpath([cwd, path])
    except ValueError:
        # On Windows, paths on different drives don't have a common path
        return path

    if common == cwd:
        return os.path.relpath(path, cwd)

    return path


def _write_locations(f: TextIO, locations: dict[tuple[str, ...], Location]) -> None:
    for path, location in locations.items():
        obj = {"path": ".".join(path), **location.as_dict()}

        json.dump(obj, f, ensure_ascii=False)
        f.write("\n")


if __name__ == "__main__":
    exit_code = main(sys.argv)

    sys.exit(exit_code)



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrOptional
from databricks.bundles.pipelines._models.postgres_catalog_config import (
    PostgresCatalogConfig,
    PostgresCatalogConfigParam,
)

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class SourceCatalogConfig:
    """
    SourceCatalogConfig contains catalog-level custom configuration parameters for each source
    """

    postgres: VariableOrOptional[PostgresCatalogConfig] = None
    """
    Postgres-specific catalog-level configuration parameters
    """

    source_catalog: VariableOrOptional[str] = None
    """
    Source catalog name
    """

    @classmethod
    def from_dict(cls, value: "SourceCatalogConfigDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "SourceCatalogConfigDict":
        return _transform_to_json_value(self)  # type:ignore


class SourceCatalogConfigDict(TypedDict, total=False):
    """"""

    postgres: VariableOrOptional[PostgresCatalogConfigParam]
    """
    Postgres-specific catalog-level configuration parameters
    """

    source_catalog: VariableOrOptional[str]
    """
    Source catalog name
    """


SourceCatalogConfigParam = SourceCatalogConfigDict | SourceCatalogConfig



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrList, VariableOrOptional
from databricks.bundles.pipelines._models.ingestion_config import (
    IngestionConfig,
    IngestionConfigParam,
)
from databricks.bundles.pipelines._models.source_config import (
    SourceConfig,
    SourceConfigParam,
)
from databricks.bundles.pipelines._models.table_specific_config import (
    TableSpecificConfig,
    TableSpecificConfigParam,
)

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class IngestionPipelineDefinition:
    """"""

    connection_name: VariableOrOptional[str] = None
    """
    Immutable. The Unity Catalog connection that this ingestion pipeline uses to communicate with the source. This is used with connectors for applications like Salesforce, Workday, and so on.
    """

    ingest_from_uc_foreign_catalog: VariableOrOptional[bool] = None
    """
    :meta private: [EXPERIMENTAL]
    
    Immutable. If set to true, the pipeline will ingest tables from the
    UC foreign catalogs directly without the need to specify a UC connection or ingestion gateway.
    The `source_catalog` fields in objects of IngestionConfig are interpreted as
    the UC foreign catalogs to ingest from.
    """

    ingestion_gateway_id: VariableOrOptional[str] = None
    """
    Immutable. Identifier for the gateway that is used by this ingestion pipeline to communicate with the source database. This is used with connectors to databases like SQL Server.
    """

    netsuite_jar_path: VariableOrOptional[str] = None
    """
    :meta private: [EXPERIMENTAL]
    """

    objects: VariableOrList[IngestionConfig] = field(default_factory=list)
    """
    Required. Settings specifying tables to replicate and the destination for the replicated tables.
    """

    source_configurations: VariableOrList[SourceConfig] = field(default_factory=list)
    """
    Top-level source configurations
    """

    table_configuration: VariableOrOptional[TableSpecificConfig] = None
    """
    Configuration settings to control the ingestion of tables. These settings are applied to all tables in the pipeline.
    """

    @classmethod
    def from_dict(cls, value: "IngestionPipelineDefinitionDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "IngestionPipelineDefinitionDict":
        return _transform_to_json_value(self)  # type:ignore


class IngestionPipelineDefinitionDict(TypedDict, total=False):
    """"""

    connection_name: VariableOrOptional[str]
    """
    Immutable. The Unity Catalog connection that this ingestion pipeline uses to communicate with the source. This is used with connectors for applications like Salesforce, Workday, and so on.
    """

    ingest_from_uc_foreign_catalog: VariableOrOptional[bool]
    """
    :meta private: [EXPERIMENTAL]
    
    Immutable. If set to true, the pipeline will ingest tables from the
    UC foreign catalogs directly without the need to specify a UC connection or ingestion gateway.
    The `source_catalog` fields in objects of IngestionConfig are interpreted as
    the UC foreign catalogs to ingest from.
    """

    ingestion_gateway_id: VariableOrOptional[str]
    """
    Immutable. Identifier for the gateway that is used by this ingestion pipeline to communicate with the source database. This is used with connectors to databases like SQL Server.
    """

    netsuite_jar_path: VariableOrOptional[str]
    """
    :meta private: [EXPERIMENTAL]
    """

    objects: VariableOrList[IngestionConfigParam]
    """
    Required. Settings specifying tables to replicate and the destination for the replicated tables.
    """

    source_configurations: VariableOrList[SourceConfigParam]
    """
    Top-level source configurations
    """

    table_configuration: VariableOrOptional[TableSpecificConfigParam]
    """
    Configuration settings to control the ingestion of tables. These settings are applied to all tables in the pipeline.
    """


IngestionPipelineDefinitionParam = (
    IngestionPipelineDefinitionDict | IngestionPipelineDefinition
)



__all__ = [
    "Lifecycle",
    "LifecycleDict",
    "LifecycleParam",
    "Schema",
    "SchemaDict",
    "SchemaGrant",
    "SchemaGrantDict",
    "SchemaGrantParam",
    "SchemaGrantPrivilege",
    "SchemaGrantPrivilegeParam",
    "SchemaParam",
]


from databricks.bundles.schemas._models.lifecycle import (
    Lifecycle,
    LifecycleDict,
    LifecycleParam,
)
from databricks.bundles.schemas._models.schema import Schema, SchemaDict, SchemaParam
from databricks.bundles.schemas._models.schema_grant import (
    SchemaGrant,
    SchemaGrantDict,
    SchemaGrantParam,
)
from databricks.bundles.schemas._models.schema_grant_privilege import (
    SchemaGrantPrivilege,
    SchemaGrantPrivilegeParam,
)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import (
    VariableOr,
    VariableOrDict,
    VariableOrOptional,
)
from databricks.bundles.jobs._models.pipeline_params import (
    PipelineParams,
    PipelineParamsParam,
)

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class RunJobTask:
    """"""

    job_id: VariableOr[int]
    """
    ID of the job to trigger.
    """

    job_parameters: VariableOrDict[str] = field(default_factory=dict)
    """
    Job-level parameters used to trigger the job.
    """

    pipeline_params: VariableOrOptional[PipelineParams] = None
    """
    Controls whether the pipeline should perform a full refresh
    """

    @classmethod
    def from_dict(cls, value: "RunJobTaskDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "RunJobTaskDict":
        return _transform_to_json_value(self)  # type:ignore


class RunJobTaskDict(TypedDict, total=False):
    """"""

    job_id: VariableOr[int]
    """
    ID of the job to trigger.
    """

    job_parameters: VariableOrDict[str]
    """
    Job-level parameters used to trigger the job.
    """

    pipeline_params: VariableOrOptional[PipelineParamsParam]
    """
    Controls whether the pipeline should perform a full refresh
    """


RunJobTaskParam = RunJobTaskDict | RunJobTask



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrList, VariableOrOptional

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class IngestionPipelineDefinitionTableSpecificConfigQueryBasedConnectorConfig:
    """
    :meta private: [EXPERIMENTAL]

    Configurations that are only applicable for query-based ingestion connectors.
    """

    cursor_columns: VariableOrList[str] = field(default_factory=list)
    """
    :meta private: [EXPERIMENTAL]
    
    The names of the monotonically increasing columns in the source table that are used to enable
    the table to be read and ingested incrementally through structured streaming.
    The columns are allowed to have repeated values but have to be non-decreasing.
    If the source data is merged into the destination (e.g., using SCD Type 1 or Type 2), these
    columns will implicitly define the `sequence_by` behavior. You can still explicitly set
    `sequence_by` to override this default.
    """

    deletion_condition: VariableOrOptional[str] = None
    """
    :meta private: [EXPERIMENTAL]
    
    Specifies a SQL WHERE condition that specifies that the source row has been deleted.
    This is sometimes referred to as "soft-deletes".
    For example: "Operation = 'DELETE'" or "is_deleted = true".
    This field is orthogonal to `hard_deletion_sync_interval_in_seconds`,
    one for soft-deletes and the other for hard-deletes.
    See also the hard_deletion_sync_min_interval_in_seconds field for
    handling of "hard deletes" where the source rows are physically removed from the table.
    """

    hard_deletion_sync_min_interval_in_seconds: VariableOrOptional[int] = None
    """
    :meta private: [EXPERIMENTAL]
    
    Specifies the minimum interval (in seconds) between snapshots on primary keys
    for detecting and synchronizing hard deletions—i.e., rows that have been
    physically removed from the source table.
    This interval acts as a lower bound. If ingestion runs less frequently than
    this value, hard deletion synchronization will align with the actual ingestion
    frequency instead of happening more often.
    If not set, hard deletion synchronization via snapshots is disabled.
    This field is mutable and can be updated without triggering a full snapshot.
    """

    @classmethod
    def from_dict(
        cls,
        value: "IngestionPipelineDefinitionTableSpecificConfigQueryBasedConnectorConfigDict",
    ) -> "Self":
        return _transform(cls, value)

    def as_dict(
        self,
    ) -> "IngestionPipelineDefinitionTableSpecificConfigQueryBasedConnectorConfigDict":
        return _transform_to_json_value(self)  # type:ignore


class IngestionPipelineDefinitionTableSpecificConfigQueryBasedConnectorConfigDict(
    TypedDict, total=False
):
    """"""

    cursor_columns: VariableOrList[str]
    """
    :meta private: [EXPERIMENTAL]
    
    The names of the monotonically increasing columns in the source table that are used to enable
    the table to be read and ingested incrementally through structured streaming.
    The columns are allowed to have repeated values but have to be non-decreasing.
    If the source data is merged into the destination (e.g., using SCD Type 1 or Type 2), these
    columns will implicitly define the `sequence_by` behavior. You can still explicitly set
    `sequence_by` to override this default.
    """

    deletion_condition: VariableOrOptional[str]
    """
    :meta private: [EXPERIMENTAL]
    
    Specifies a SQL WHERE condition that specifies that the source row has been deleted.
    This is sometimes referred to as "soft-deletes".
    For example: "Operation = 'DELETE'" or "is_deleted = true".
    This field is orthogonal to `hard_deletion_sync_interval_in_seconds`,
    one for soft-deletes and the other for hard-deletes.
    See also the hard_deletion_sync_min_interval_in_seconds field for
    handling of "hard deletes" where the source rows are physically removed from the table.
    """

    hard_deletion_sync_min_interval_in_seconds: VariableOrOptional[int]
    """
    :meta private: [EXPERIMENTAL]
    
    Specifies the minimum interval (in seconds) between snapshots on primary keys
    for detecting and synchronizing hard deletions—i.e., rows that have been
    physically removed from the source table.
    This interval acts as a lower bound. If ingestion runs less frequently than
    this value, hard deletion synchronization will align with the actual ingestion
    frequency instead of happening more often.
    If not set, hard deletion synchronization via snapshots is disabled.
    This field is mutable and can be updated without triggering a full snapshot.
    """


IngestionPipelineDefinitionTableSpecificConfigQueryBasedConnectorConfigParam = (
    IngestionPipelineDefinitionTableSpecificConfigQueryBasedConnectorConfigDict
    | IngestionPipelineDefinitionTableSpecificConfigQueryBasedConnectorConfig
)



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOr, VariableOrOptional

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class RCranLibrary:
    """"""

    package: VariableOr[str]
    """
    The name of the CRAN package to install.
    """

    repo: VariableOrOptional[str] = None
    """
    The repository where the package can be found. If not specified, the default CRAN repo is used.
    """

    @classmethod
    def from_dict(cls, value: "RCranLibraryDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "RCranLibraryDict":
        return _transform_to_json_value(self)  # type:ignore


class RCranLibraryDict(TypedDict, total=False):
    """"""

    package: VariableOr[str]
    """
    The name of the CRAN package to install.
    """

    repo: VariableOrOptional[str]
    """
    The repository where the package can be found. If not specified, the default CRAN repo is used.
    """


RCranLibraryParam = RCranLibraryDict | RCranLibrary



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrOptional
from databricks.bundles.jobs._models.storage_mode import StorageMode, StorageModeParam

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class PowerBiTable:
    """"""

    catalog: VariableOrOptional[str] = None
    """
    The catalog name in Databricks
    """

    name: VariableOrOptional[str] = None
    """
    The table name in Databricks
    """

    schema: VariableOrOptional[str] = None
    """
    The schema name in Databricks
    """

    storage_mode: VariableOrOptional[StorageMode] = None
    """
    The Power BI storage mode of the table
    """

    @classmethod
    def from_dict(cls, value: "PowerBiTableDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "PowerBiTableDict":
        return _transform_to_json_value(self)  # type:ignore


class PowerBiTableDict(TypedDict, total=False):
    """"""

    catalog: VariableOrOptional[str]
    """
    The catalog name in Databricks
    """

    name: VariableOrOptional[str]
    """
    The table name in Databricks
    """

    schema: VariableOrOptional[str]
    """
    The schema name in Databricks
    """

    storage_mode: VariableOrOptional[StorageModeParam]
    """
    The Power BI storage mode of the table
    """


PowerBiTableParam = PowerBiTableDict | PowerBiTable



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import (
    VariableOr,
    VariableOrList,
    VariableOrOptional,
)
from databricks.bundles.pipelines._models.day_of_week import DayOfWeek, DayOfWeekParam

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class RestartWindow:
    """
    :meta private: [EXPERIMENTAL]
    """

    start_hour: VariableOr[int]
    """
    An integer between 0 and 23 denoting the start hour for the restart window in the 24-hour day.
    Continuous pipeline restart is triggered only within a five-hour window starting at this hour.
    """

    days_of_week: VariableOrList[DayOfWeek] = field(default_factory=list)
    """
    Days of week in which the restart is allowed to happen (within a five-hour window starting at start_hour).
    If not specified all days of the week will be used.
    """

    time_zone_id: VariableOrOptional[str] = None
    """
    Time zone id of restart window. See https://docs.databricks.com/sql/language-manual/sql-ref-syntax-aux-conf-mgmt-set-timezone.html for details.
    If not specified, UTC will be used.
    """

    @classmethod
    def from_dict(cls, value: "RestartWindowDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "RestartWindowDict":
        return _transform_to_json_value(self)  # type:ignore


class RestartWindowDict(TypedDict, total=False):
    """"""

    start_hour: VariableOr[int]
    """
    An integer between 0 and 23 denoting the start hour for the restart window in the 24-hour day.
    Continuous pipeline restart is triggered only within a five-hour window starting at this hour.
    """

    days_of_week: VariableOrList[DayOfWeekParam]
    """
    Days of week in which the restart is allowed to happen (within a five-hour window starting at start_hour).
    If not specified all days of the week will be used.
    """

    time_zone_id: VariableOrOptional[str]
    """
    Time zone id of restart window. See https://docs.databricks.com/sql/language-manual/sql-ref-syntax-aux-conf-mgmt-set-timezone.html for details.
    If not specified, UTC will be used.
    """


RestartWindowParam = RestartWindowDict | RestartWindow



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrOptional

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class AutoScale:
    """"""

    max_workers: VariableOrOptional[int] = None
    """
    The maximum number of workers to which the cluster can scale up when overloaded.
    Note that `max_workers` must be strictly greater than `min_workers`.
    """

    min_workers: VariableOrOptional[int] = None
    """
    The minimum number of workers to which the cluster can scale down when underutilized.
    It is also the initial number of workers the cluster will have after creation.
    """

    @classmethod
    def from_dict(cls, value: "AutoScaleDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "AutoScaleDict":
        return _transform_to_json_value(self)  # type:ignore


class AutoScaleDict(TypedDict, total=False):
    """"""

    max_workers: VariableOrOptional[int]
    """
    The maximum number of workers to which the cluster can scale up when overloaded.
    Note that `max_workers` must be strictly greater than `min_workers`.
    """

    min_workers: VariableOrOptional[int]
    """
    The minimum number of workers to which the cluster can scale down when underutilized.
    It is also the initial number of workers the cluster will have after creation.
    """


AutoScaleParam = AutoScaleDict | AutoScale



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import (
    VariableOr,
    VariableOrList,
    VariableOrOptional,
)

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class SparkJarTask:
    """"""

    main_class_name: VariableOr[str]
    """
    The full name of the class containing the main method to be executed. This class must be contained in a JAR provided as a library.
    
    The code must use `SparkContext.getOrCreate` to obtain a Spark context; otherwise, runs of the job fail.
    """

    jar_uri: VariableOrOptional[str] = None
    """
    [DEPRECATED] Deprecated since 04/2016. For classic compute, provide a `jar` through the `libraries` field instead. For serverless compute, provide a `jar` though the `java_dependencies` field inside the `environments` list.
    
    See the examples of classic and serverless compute usage at the top of the page.
    """

    parameters: VariableOrList[str] = field(default_factory=list)
    """
    Parameters passed to the main method.
    
    Use [Task parameter variables](https://docs.databricks.com/jobs.html#parameter-variables) to set parameters containing information about job runs.
    """

    run_as_repl: VariableOrOptional[bool] = None
    """
    [DEPRECATED] Deprecated. A value of `false` is no longer supported.
    """

    @classmethod
    def from_dict(cls, value: "SparkJarTaskDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "SparkJarTaskDict":
        return _transform_to_json_value(self)  # type:ignore


class SparkJarTaskDict(TypedDict, total=False):
    """"""

    main_class_name: VariableOr[str]
    """
    The full name of the class containing the main method to be executed. This class must be contained in a JAR provided as a library.
    
    The code must use `SparkContext.getOrCreate` to obtain a Spark context; otherwise, runs of the job fail.
    """

    jar_uri: VariableOrOptional[str]
    """
    [DEPRECATED] Deprecated since 04/2016. For classic compute, provide a `jar` through the `libraries` field instead. For serverless compute, provide a `jar` though the `java_dependencies` field inside the `environments` list.
    
    See the examples of classic and serverless compute usage at the top of the page.
    """

    parameters: VariableOrList[str]
    """
    Parameters passed to the main method.
    
    Use [Task parameter variables](https://docs.databricks.com/jobs.html#parameter-variables) to set parameters containing information about job runs.
    """

    run_as_repl: VariableOrOptional[bool]
    """
    [DEPRECATED] Deprecated. A value of `false` is no longer supported.
    """


SparkJarTaskParam = SparkJarTaskDict | SparkJarTask



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrOptional
from databricks.bundles.pipelines._models.report_spec import ReportSpec, ReportSpecParam
from databricks.bundles.pipelines._models.schema_spec import SchemaSpec, SchemaSpecParam
from databricks.bundles.pipelines._models.table_spec import TableSpec, TableSpecParam

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class IngestionConfig:
    """"""

    report: VariableOrOptional[ReportSpec] = None
    """
    Select a specific source report.
    """

    schema: VariableOrOptional[SchemaSpec] = None
    """
    Select all tables from a specific source schema.
    """

    table: VariableOrOptional[TableSpec] = None
    """
    Select a specific source table.
    """

    @classmethod
    def from_dict(cls, value: "IngestionConfigDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "IngestionConfigDict":
        return _transform_to_json_value(self)  # type:ignore


class IngestionConfigDict(TypedDict, total=False):
    """"""

    report: VariableOrOptional[ReportSpecParam]
    """
    Select a specific source report.
    """

    schema: VariableOrOptional[SchemaSpecParam]
    """
    Select all tables from a specific source schema.
    """

    table: VariableOrOptional[TableSpecParam]
    """
    Select a specific source table.
    """


IngestionConfigParam = IngestionConfigDict | IngestionConfig



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOr, VariableOrOptional
from databricks.bundles.jobs._models.pause_status import PauseStatus, PauseStatusParam

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class CronSchedule:
    """"""

    quartz_cron_expression: VariableOr[str]
    """
    A Cron expression using Quartz syntax that describes the schedule for a job. See [Cron Trigger](http://www.quartz-scheduler.org/documentation/quartz-2.3.0/tutorials/crontrigger.html) for details. This field is required.
    """

    timezone_id: VariableOr[str]
    """
    A Java timezone ID. The schedule for a job is resolved with respect to this timezone. See [Java TimeZone](https://docs.oracle.com/javase/7/docs/api/java/util/TimeZone.html) for details. This field is required.
    """

    pause_status: VariableOrOptional[PauseStatus] = None
    """
    Indicate whether this schedule is paused or not.
    """

    @classmethod
    def from_dict(cls, value: "CronScheduleDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "CronScheduleDict":
        return _transform_to_json_value(self)  # type:ignore


class CronScheduleDict(TypedDict, total=False):
    """"""

    quartz_cron_expression: VariableOr[str]
    """
    A Cron expression using Quartz syntax that describes the schedule for a job. See [Cron Trigger](http://www.quartz-scheduler.org/documentation/quartz-2.3.0/tutorials/crontrigger.html) for details. This field is required.
    """

    timezone_id: VariableOr[str]
    """
    A Java timezone ID. The schedule for a job is resolved with respect to this timezone. See [Java TimeZone](https://docs.oracle.com/javase/7/docs/api/java/util/TimeZone.html) for details. This field is required.
    """

    pause_status: VariableOrOptional[PauseStatusParam]
    """
    Indicate whether this schedule is paused or not.
    """


CronScheduleParam = CronScheduleDict | CronSchedule



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import (
    VariableOr,
    VariableOrDict,
    VariableOrOptional,
)

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class CleanRoomsNotebookTask:
    """
    Clean Rooms notebook task for V1 Clean Room service (GA).
    Replaces the deprecated CleanRoomNotebookTask (defined above) which was for V0 service.
    """

    clean_room_name: VariableOr[str]
    """
    The clean room that the notebook belongs to.
    """

    notebook_name: VariableOr[str]
    """
    Name of the notebook being run.
    """

    etag: VariableOrOptional[str] = None
    """
    Checksum to validate the freshness of the notebook resource (i.e. the notebook being run is the latest version).
    It can be fetched by calling the :method:cleanroomassets/get API.
    """

    notebook_base_parameters: VariableOrDict[str] = field(default_factory=dict)
    """
    Base parameters to be used for the clean room notebook job.
    """

    @classmethod
    def from_dict(cls, value: "CleanRoomsNotebookTaskDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "CleanRoomsNotebookTaskDict":
        return _transform_to_json_value(self)  # type:ignore


class CleanRoomsNotebookTaskDict(TypedDict, total=False):
    """"""

    clean_room_name: VariableOr[str]
    """
    The clean room that the notebook belongs to.
    """

    notebook_name: VariableOr[str]
    """
    Name of the notebook being run.
    """

    etag: VariableOrOptional[str]
    """
    Checksum to validate the freshness of the notebook resource (i.e. the notebook being run is the latest version).
    It can be fetched by calling the :method:cleanroomassets/get API.
    """

    notebook_base_parameters: VariableOrDict[str]
    """
    Base parameters to be used for the clean room notebook job.
    """


CleanRoomsNotebookTaskParam = CleanRoomsNotebookTaskDict | CleanRoomsNotebookTask



# See: https://packaging.python.org/guides/packaging-namespace-packages/#pkgutil-style-namespace-packages
#
# This file must only contain the following line, or other packages in the databricks.* namespace
# may not be importable. The contents of this file must be byte-for-byte equivalent across all packages.
# If they are not, parallel package installation may lead to clobbered and invalid files.
# Also see https://github.com/databricks/databricks-sdk-py/issues/343.
__path__ = __import__("pkgutil").extend_path(__path__, __name__)



__all__ = [
    "Adlsgen2Info",
    "Adlsgen2InfoDict",
    "Adlsgen2InfoParam",
    "AuthenticationMethod",
    "AuthenticationMethodParam",
    "AutoScale",
    "AutoScaleDict",
    "AutoScaleParam",
    "AwsAttributes",
    "AwsAttributesDict",
    "AwsAttributesParam",
    "AwsAvailability",
    "AwsAvailabilityParam",
    "AzureAttributes",
    "AzureAttributesDict",
    "AzureAttributesParam",
    "AzureAvailability",
    "AzureAvailabilityParam",
    "CleanRoomsNotebookTask",
    "CleanRoomsNotebookTaskDict",
    "CleanRoomsNotebookTaskParam",
    "ClientsTypes",
    "ClientsTypesDict",
    "ClientsTypesParam",
    "ClusterLogConf",
    "ClusterLogConfDict",
    "ClusterLogConfParam",
    "ClusterSpec",
    "ClusterSpecDict",
    "ClusterSpecParam",
    "ComputeConfig",
    "ComputeConfigDict",
    "ComputeConfigParam",
    "Condition",
    "ConditionParam",
    "ConditionTask",
    "ConditionTaskDict",
    "ConditionTaskOp",
    "ConditionTaskOpParam",
    "ConditionTaskParam",
    "Continuous",
    "ContinuousDict",
    "ContinuousParam",
    "CronSchedule",
    "CronScheduleDict",
    "CronScheduleParam",
    "DashboardTask",
    "DashboardTaskDict",
    "DashboardTaskParam",
    "DataSecurityMode",
    "DataSecurityModeParam",
    "DbfsStorageInfo",
    "DbfsStorageInfoDict",
    "DbfsStorageInfoParam",
    "DbtPlatformTask",
    "DbtPlatformTaskDict",
    "DbtPlatformTaskParam",
    "DbtTask",
    "DbtTaskDict",
    "DbtTaskParam",
    "DockerBasicAuth",
    "DockerBasicAuthDict",
    "DockerBasicAuthParam",
    "DockerImage",
    "DockerImageDict",
    "DockerImageParam",
    "EbsVolumeType",
    "EbsVolumeTypeParam",
    "Environment",
    "EnvironmentDict",
    "EnvironmentParam",
    "FileArrivalTriggerConfiguration",
    "FileArrivalTriggerConfigurationDict",
    "FileArrivalTriggerConfigurationParam",
    "ForEachTask",
    "ForEachTaskDict",
    "ForEachTaskParam",
    "GcpAttributes",
    "GcpAttributesDict",
    "GcpAttributesParam",
    "GcpAvailability",
    "GcpAvailabilityParam",
    "GcsStorageInfo",
    "GcsStorageInfoDict",
    "GcsStorageInfoParam",
    "GenAiComputeTask",
    "GenAiComputeTaskDict",
    "GenAiComputeTaskParam",
    "GitProvider",
    "GitProviderParam",
    "GitSource",
    "GitSourceDict",
    "GitSourceParam",
    "InitScriptInfo",
    "InitScriptInfoDict",
    "InitScriptInfoParam",
    "Job",
    "JobCluster",
    "JobClusterDict",
    "JobClusterParam",
    "JobDict",
    "JobEmailNotifications",
    "JobEmailNotificationsDict",
    "JobEmailNotificationsParam",
    "JobEnvironment",
    "JobEnvironmentDict",
    "JobEnvironmentParam",
    "JobNotificationSettings",
    "JobNotificationSettingsDict",
    "JobNotificationSettingsParam",
    "JobParam",
    "JobParameterDefinition",
    "JobParameterDefinitionDict",
    "JobParameterDefinitionParam",
    "JobPermission",
    "JobPermissionDict",
    "JobPermissionLevel",
    "JobPermissionLevelParam",
    "JobPermissionParam",
    "JobRunAs",
    "JobRunAsDict",
    "JobRunAsParam",
    "JobsHealthMetric",
    "JobsHealthMetricParam",
    "JobsHealthOperator",
    "JobsHealthOperatorParam",
    "JobsHealthRule",
    "JobsHealthRuleDict",
    "JobsHealthRuleParam",
    "JobsHealthRules",
    "JobsHealthRulesDict",
    "JobsHealthRulesParam",
    "Kind",
    "KindParam",
    "Library",
    "LibraryDict",
    "LibraryParam",
    "Lifecycle",
    "LifecycleDict",
    "LifecycleParam",
    "LocalFileInfo",
    "LocalFileInfoDict",
    "LocalFileInfoParam",
    "LogAnalyticsInfo",
    "LogAnalyticsInfoDict",
    "LogAnalyticsInfoParam",
    "MavenLibrary",
    "MavenLibraryDict",
    "MavenLibraryParam",
    "ModelTriggerConfiguration",
    "ModelTriggerConfigurationCondition",
    "ModelTriggerConfigurationConditionParam",
    "ModelTriggerConfigurationDict",
    "ModelTriggerConfigurationParam",
    "NotebookTask",
    "NotebookTaskDict",
    "NotebookTaskParam",
    "PauseStatus",
    "PauseStatusParam",
    "PerformanceTarget",
    "PerformanceTargetParam",
    "PeriodicTriggerConfiguration",
    "PeriodicTriggerConfigurationDict",
    "PeriodicTriggerConfigurationParam",
    "PeriodicTriggerConfigurationTimeUnit",
    "PeriodicTriggerConfigurationTimeUnitParam",
    "PipelineParams",
    "PipelineParamsDict",
    "PipelineParamsParam",
    "PipelineTask",
    "PipelineTaskDict",
    "PipelineTaskParam",
    "PowerBiModel",
    "PowerBiModelDict",
    "PowerBiModelParam",
    "PowerBiTable",
    "PowerBiTableDict",
    "PowerBiTableParam",
    "PowerBiTask",
    "PowerBiTaskDict",
    "PowerBiTaskParam",
    "PythonPyPiLibrary",
    "PythonPyPiLibraryDict",
    "PythonPyPiLibraryParam",
    "PythonWheelTask",
    "PythonWheelTaskDict",
    "PythonWheelTaskParam",
    "QueueSettings",
    "QueueSettingsDict",
    "QueueSettingsParam",
    "RCranLibrary",
    "RCranLibraryDict",
    "RCranLibraryParam",
    "RunIf",
    "RunIfParam",
    "RunJobTask",
    "RunJobTaskDict",
    "RunJobTaskParam",
    "RuntimeEngine",
    "RuntimeEngineParam",
    "S3StorageInfo",
    "S3StorageInfoDict",
    "S3StorageInfoParam",
    "Source",
    "SourceParam",
    "SparkJarTask",
    "SparkJarTaskDict",
    "SparkJarTaskParam",
    "SparkPythonTask",
    "SparkPythonTaskDict",
    "SparkPythonTaskParam",
    "SparkSubmitTask",
    "SparkSubmitTaskDict",
    "SparkSubmitTaskParam",
    "SqlTask",
    "SqlTaskAlert",
    "SqlTaskAlertDict",
    "SqlTaskAlertParam",
    "SqlTaskDashboard",
    "SqlTaskDashboardDict",
    "SqlTaskDashboardParam",
    "SqlTaskDict",
    "SqlTaskFile",
    "SqlTaskFileDict",
    "SqlTaskFileParam",
    "SqlTaskParam",
    "SqlTaskQuery",
    "SqlTaskQueryDict",
    "SqlTaskQueryParam",
    "SqlTaskSubscription",
    "SqlTaskSubscriptionDict",
    "SqlTaskSubscriptionParam",
    "StorageMode",
    "StorageModeParam",
    "Subscription",
    "SubscriptionDict",
    "SubscriptionParam",
    "SubscriptionSubscriber",
    "SubscriptionSubscriberDict",
    "SubscriptionSubscriberParam",
    "TableUpdateTriggerConfiguration",
    "TableUpdateTriggerConfigurationDict",
    "TableUpdateTriggerConfigurationParam",
    "Task",
    "TaskDependency",
    "TaskDependencyDict",
    "TaskDependencyParam",
    "TaskDict",
    "TaskEmailNotifications",
    "TaskEmailNotificationsDict",
    "TaskEmailNotificationsParam",
    "TaskNotificationSettings",
    "TaskNotificationSettingsDict",
    "TaskNotificationSettingsParam",
    "TaskParam",
    "TaskRetryMode",
    "TaskRetryModeParam",
    "TriggerSettings",
    "TriggerSettingsDict",
    "TriggerSettingsParam",
    "VolumesStorageInfo",
    "VolumesStorageInfoDict",
    "VolumesStorageInfoParam",
    "Webhook",
    "WebhookDict",
    "WebhookNotifications",
    "WebhookNotificationsDict",
    "WebhookNotificationsParam",
    "WebhookParam",
    "WorkloadType",
    "WorkloadTypeDict",
    "WorkloadTypeParam",
    "WorkspaceStorageInfo",
    "WorkspaceStorageInfoDict",
    "WorkspaceStorageInfoParam",
]


from databricks.bundles.jobs._models.adlsgen2_info import (
    Adlsgen2Info,
    Adlsgen2InfoDict,
    Adlsgen2InfoParam,
)
from databricks.bundles.jobs._models.authentication_method import (
    AuthenticationMethod,
    AuthenticationMethodParam,
)
from databricks.bundles.jobs._models.auto_scale import (
    AutoScale,
    AutoScaleDict,
    AutoScaleParam,
)
from databricks.bundles.jobs._models.aws_attributes import (
    AwsAttributes,
    AwsAttributesDict,
    AwsAttributesParam,
)
from databricks.bundles.jobs._models.aws_availability import (
    AwsAvailability,
    AwsAvailabilityParam,
)
from databricks.bundles.jobs._models.azure_attributes import (
    AzureAttributes,
    AzureAttributesDict,
    AzureAttributesParam,
)
from databricks.bundles.jobs._models.azure_availability import (
    AzureAvailability,
    AzureAvailabilityParam,
)
from databricks.bundles.jobs._models.clean_rooms_notebook_task import (
    CleanRoomsNotebookTask,
    CleanRoomsNotebookTaskDict,
    CleanRoomsNotebookTaskParam,
)
from databricks.bundles.jobs._models.clients_types import (
    ClientsTypes,
    ClientsTypesDict,
    ClientsTypesParam,
)
from databricks.bundles.jobs._models.cluster_log_conf import (
    ClusterLogConf,
    ClusterLogConfDict,
    ClusterLogConfParam,
)
from databricks.bundles.jobs._models.cluster_spec import (
    ClusterSpec,
    ClusterSpecDict,
    ClusterSpecParam,
)
from databricks.bundles.jobs._models.compute_config import (
    ComputeConfig,
    ComputeConfigDict,
    ComputeConfigParam,
)
from databricks.bundles.jobs._models.condition import Condition, ConditionParam
from databricks.bundles.jobs._models.condition_task import (
    ConditionTask,
    ConditionTaskDict,
    ConditionTaskParam,
)
from databricks.bundles.jobs._models.condition_task_op import (
    ConditionTaskOp,
    ConditionTaskOpParam,
)
from databricks.bundles.jobs._models.continuous import (
    Continuous,
    ContinuousDict,
    ContinuousParam,
)
from databricks.bundles.jobs._models.cron_schedule import (
    CronSchedule,
    CronScheduleDict,
    CronScheduleParam,
)
from databricks.bundles.jobs._models.dashboard_task import (
    DashboardTask,
    DashboardTaskDict,
    DashboardTaskParam,
)
from databricks.bundles.jobs._models.data_security_mode import (
    DataSecurityMode,
    DataSecurityModeParam,
)
from databricks.bundles.jobs._models.dbfs_storage_info import (
    DbfsStorageInfo,
    DbfsStorageInfoDict,
    DbfsStorageInfoParam,
)
from databricks.bundles.jobs._models.dbt_platform_task import (
    DbtPlatformTask,
    DbtPlatformTaskDict,
    DbtPlatformTaskParam,
)
from databricks.bundles.jobs._models.dbt_task import DbtTask, DbtTaskDict, DbtTaskParam
from databricks.bundles.jobs._models.docker_basic_auth import (
    DockerBasicAuth,
    DockerBasicAuthDict,
    DockerBasicAuthParam,
)
from databricks.bundles.jobs._models.docker_image import (
    DockerImage,
    DockerImageDict,
    DockerImageParam,
)
from databricks.bundles.jobs._models.ebs_volume_type import (
    EbsVolumeType,
    EbsVolumeTypeParam,
)
from databricks.bundles.jobs._models.environment import (
    Environment,
    EnvironmentDict,
    EnvironmentParam,
)
from databricks.bundles.jobs._models.file_arrival_trigger_configuration import (
    FileArrivalTriggerConfiguration,
    FileArrivalTriggerConfigurationDict,
    FileArrivalTriggerConfigurationParam,
)
from databricks.bundles.jobs._models.for_each_task import (
    ForEachTask,
    ForEachTaskDict,
    ForEachTaskParam,
)
from databricks.bundles.jobs._models.gcp_attributes import (
    GcpAttributes,
    GcpAttributesDict,
    GcpAttributesParam,
)
from databricks.bundles.jobs._models.gcp_availability import (
    GcpAvailability,
    GcpAvailabilityParam,
)
from databricks.bundles.jobs._models.gcs_storage_info import (
    GcsStorageInfo,
    GcsStorageInfoDict,
    GcsStorageInfoParam,
)
from databricks.bundles.jobs._models.gen_ai_compute_task import (
    GenAiComputeTask,
    GenAiComputeTaskDict,
    GenAiComputeTaskParam,
)
from databricks.bundles.jobs._models.git_provider import GitProvider, GitProviderParam
from databricks.bundles.jobs._models.git_source import (
    GitSource,
    GitSourceDict,
    GitSourceParam,
)
from databricks.bundles.jobs._models.init_script_info import (
    InitScriptInfo,
    InitScriptInfoDict,
    InitScriptInfoParam,
)
from databricks.bundles.jobs._models.job import Job, JobDict, JobParam
from databricks.bundles.jobs._models.job_cluster import (
    JobCluster,
    JobClusterDict,
    JobClusterParam,
)
from databricks.bundles.jobs._models.job_email_notifications import (
    JobEmailNotifications,
    JobEmailNotificationsDict,
    JobEmailNotificationsParam,
)
from databricks.bundles.jobs._models.job_environment import (
    JobEnvironment,
    JobEnvironmentDict,
    JobEnvironmentParam,
)
from databricks.bundles.jobs._models.job_notification_settings import (
    JobNotificationSettings,
    JobNotificationSettingsDict,
    JobNotificationSettingsParam,
)
from databricks.bundles.jobs._models.job_parameter_definition import (
    JobParameterDefinition,
    JobParameterDefinitionDict,
    JobParameterDefinitionParam,
)
from databricks.bundles.jobs._models.job_permission import (
    JobPermission,
    JobPermissionDict,
    JobPermissionParam,
)
from databricks.bundles.jobs._models.job_permission_level import (
    JobPermissionLevel,
    JobPermissionLevelParam,
)
from databricks.bundles.jobs._models.job_run_as import (
    JobRunAs,
    JobRunAsDict,
    JobRunAsParam,
)
from databricks.bundles.jobs._models.jobs_health_metric import (
    JobsHealthMetric,
    JobsHealthMetricParam,
)
from databricks.bundles.jobs._models.jobs_health_operator import (
    JobsHealthOperator,
    JobsHealthOperatorParam,
)
from databricks.bundles.jobs._models.jobs_health_rule import (
    JobsHealthRule,
    JobsHealthRuleDict,
    JobsHealthRuleParam,
)
from databricks.bundles.jobs._models.jobs_health_rules import (
    JobsHealthRules,
    JobsHealthRulesDict,
    JobsHealthRulesParam,
)
from databricks.bundles.jobs._models.kind import Kind, KindParam
from databricks.bundles.jobs._models.library import Library, LibraryDict, LibraryParam
from databricks.bundles.jobs._models.lifecycle import (
    Lifecycle,
    LifecycleDict,
    LifecycleParam,
)
from databricks.bundles.jobs._models.local_file_info import (
    LocalFileInfo,
    LocalFileInfoDict,
    LocalFileInfoParam,
)
from databricks.bundles.jobs._models.log_analytics_info import (
    LogAnalyticsInfo,
    LogAnalyticsInfoDict,
    LogAnalyticsInfoParam,
)
from databricks.bundles.jobs._models.maven_library import (
    MavenLibrary,
    MavenLibraryDict,
    MavenLibraryParam,
)
from databricks.bundles.jobs._models.model_trigger_configuration import (
    ModelTriggerConfiguration,
    ModelTriggerConfigurationDict,
    ModelTriggerConfigurationParam,
)
from databricks.bundles.jobs._models.model_trigger_configuration_condition import (
    ModelTriggerConfigurationCondition,
    ModelTriggerConfigurationConditionParam,
)
from databricks.bundles.jobs._models.notebook_task import (
    NotebookTask,
    NotebookTaskDict,
    NotebookTaskParam,
)
from databricks.bundles.jobs._models.pause_status import PauseStatus, PauseStatusParam
from databricks.bundles.jobs._models.performance_target import (
    PerformanceTarget,
    PerformanceTargetParam,
)
from databricks.bundles.jobs._models.periodic_trigger_configuration import (
    PeriodicTriggerConfiguration,
    PeriodicTriggerConfigurationDict,
    PeriodicTriggerConfigurationParam,
)
from databricks.bundles.jobs._models.periodic_trigger_configuration_time_unit import (
    PeriodicTriggerConfigurationTimeUnit,
    PeriodicTriggerConfigurationTimeUnitParam,
)
from databricks.bundles.jobs._models.pipeline_params import (
    PipelineParams,
    PipelineParamsDict,
    PipelineParamsParam,
)
from databricks.bundles.jobs._models.pipeline_task import (
    PipelineTask,
    PipelineTaskDict,
    PipelineTaskParam,
)
from databricks.bundles.jobs._models.power_bi_model import (
    PowerBiModel,
    PowerBiModelDict,
    PowerBiModelParam,
)
from databricks.bundles.jobs._models.power_bi_table import (
    PowerBiTable,
    PowerBiTableDict,
    PowerBiTableParam,
)
from databricks.bundles.jobs._models.power_bi_task import (
    PowerBiTask,
    PowerBiTaskDict,
    PowerBiTaskParam,
)
from databricks.bundles.jobs._models.python_py_pi_library import (
    PythonPyPiLibrary,
    PythonPyPiLibraryDict,
    PythonPyPiLibraryParam,
)
from databricks.bundles.jobs._models.python_wheel_task import (
    PythonWheelTask,
    PythonWheelTaskDict,
    PythonWheelTaskParam,
)
from databricks.bundles.jobs._models.queue_settings import (
    QueueSettings,
    QueueSettingsDict,
    QueueSettingsParam,
)
from databricks.bundles.jobs._models.r_cran_library import (
    RCranLibrary,
    RCranLibraryDict,
    RCranLibraryParam,
)
from databricks.bundles.jobs._models.run_if import RunIf, RunIfParam
from databricks.bundles.jobs._models.run_job_task import (
    RunJobTask,
    RunJobTaskDict,
    RunJobTaskParam,
)
from databricks.bundles.jobs._models.runtime_engine import (
    RuntimeEngine,
    RuntimeEngineParam,
)
from databricks.bundles.jobs._models.s3_storage_info import (
    S3StorageInfo,
    S3StorageInfoDict,
    S3StorageInfoParam,
)
from databricks.bundles.jobs._models.source import Source, SourceParam
from databricks.bundles.jobs._models.spark_jar_task import (
    SparkJarTask,
    SparkJarTaskDict,
    SparkJarTaskParam,
)
from databricks.bundles.jobs._models.spark_python_task import (
    SparkPythonTask,
    SparkPythonTaskDict,
    SparkPythonTaskParam,
)
from databricks.bundles.jobs._models.spark_submit_task import (
    SparkSubmitTask,
    SparkSubmitTaskDict,
    SparkSubmitTaskParam,
)
from databricks.bundles.jobs._models.sql_task import SqlTask, SqlTaskDict, SqlTaskParam
from databricks.bundles.jobs._models.sql_task_alert import (
    SqlTaskAlert,
    SqlTaskAlertDict,
    SqlTaskAlertParam,
)
from databricks.bundles.jobs._models.sql_task_dashboard import (
    SqlTaskDashboard,
    SqlTaskDashboardDict,
    SqlTaskDashboardParam,
)
from databricks.bundles.jobs._models.sql_task_file import (
    SqlTaskFile,
    SqlTaskFileDict,
    SqlTaskFileParam,
)
from databricks.bundles.jobs._models.sql_task_query import (
    SqlTaskQuery,
    SqlTaskQueryDict,
    SqlTaskQueryParam,
)
from databricks.bundles.jobs._models.sql_task_subscription import (
    SqlTaskSubscription,
    SqlTaskSubscriptionDict,
    SqlTaskSubscriptionParam,
)
from databricks.bundles.jobs._models.storage_mode import StorageMode, StorageModeParam
from databricks.bundles.jobs._models.subscription import (
    Subscription,
    SubscriptionDict,
    SubscriptionParam,
)
from databricks.bundles.jobs._models.subscription_subscriber import (
    SubscriptionSubscriber,
    SubscriptionSubscriberDict,
    SubscriptionSubscriberParam,
)
from databricks.bundles.jobs._models.table_update_trigger_configuration import (
    TableUpdateTriggerConfiguration,
    TableUpdateTriggerConfigurationDict,
    TableUpdateTriggerConfigurationParam,
)
from databricks.bundles.jobs._models.task import Task, TaskDict, TaskParam
from databricks.bundles.jobs._models.task_dependency import (
    TaskDependency,
    TaskDependencyDict,
    TaskDependencyParam,
)
from databricks.bundles.jobs._models.task_email_notifications import (
    TaskEmailNotifications,
    TaskEmailNotificationsDict,
    TaskEmailNotificationsParam,
)
from databricks.bundles.jobs._models.task_notification_settings import (
    TaskNotificationSettings,
    TaskNotificationSettingsDict,
    TaskNotificationSettingsParam,
)
from databricks.bundles.jobs._models.task_retry_mode import (
    TaskRetryMode,
    TaskRetryModeParam,
)
from databricks.bundles.jobs._models.trigger_settings import (
    TriggerSettings,
    TriggerSettingsDict,
    TriggerSettingsParam,
)
from databricks.bundles.jobs._models.volumes_storage_info import (
    VolumesStorageInfo,
    VolumesStorageInfoDict,
    VolumesStorageInfoParam,
)
from databricks.bundles.jobs._models.webhook import Webhook, WebhookDict, WebhookParam
from databricks.bundles.jobs._models.webhook_notifications import (
    WebhookNotifications,
    WebhookNotificationsDict,
    WebhookNotificationsParam,
)
from databricks.bundles.jobs._models.workload_type import (
    WorkloadType,
    WorkloadTypeDict,
    WorkloadTypeParam,
)
from databricks.bundles.jobs._models.workspace_storage_info import (
    WorkspaceStorageInfo,
    WorkspaceStorageInfoDict,
    WorkspaceStorageInfoParam,
)


def _resolve_recursive_imports():
    import typing

    from databricks.bundles.core._variable import VariableOr
    from databricks.bundles.jobs._models.task import Task

    ForEachTask.__annotations__ = typing.get_type_hints(
        ForEachTask,
        globalns={"Task": Task, "VariableOr": VariableOr},
    )


_resolve_recursive_imports()



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrList
from databricks.bundles.jobs._models.jobs_health_rule import (
    JobsHealthRule,
    JobsHealthRuleParam,
)

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class JobsHealthRules:
    """
    An optional set of health rules that can be defined for this job.
    """

    rules: VariableOrList[JobsHealthRule] = field(default_factory=list)

    @classmethod
    def from_dict(cls, value: "JobsHealthRulesDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "JobsHealthRulesDict":
        return _transform_to_json_value(self)  # type:ignore


class JobsHealthRulesDict(TypedDict, total=False):
    """"""

    rules: VariableOrList[JobsHealthRuleParam]


JobsHealthRulesParam = JobsHealthRulesDict | JobsHealthRules



from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOrOptional

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class Lifecycle:
    """"""

    prevent_destroy: VariableOrOptional[bool] = None
    """
    Lifecycle setting to prevent the resource from being destroyed.
    """

    @classmethod
    def from_dict(cls, value: "LifecycleDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "LifecycleDict":
        return _transform_to_json_value(self)  # type:ignore


class LifecycleDict(TypedDict, total=False):
    """"""

    prevent_destroy: VariableOrOptional[bool]
    """
    Lifecycle setting to prevent the resource from being destroyed.
    """


LifecycleParam = LifecycleDict | Lifecycle



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from databricks.bundles.core._transform import _transform
from databricks.bundles.core._transform_to_json import _transform_to_json_value
from databricks.bundles.core._variable import VariableOr, VariableOrList
from databricks.bundles.volumes._models.volume_grant_privilege import (
    VolumeGrantPrivilege,
    VolumeGrantPrivilegeParam,
)

if TYPE_CHECKING:
    from typing_extensions import Self


@dataclass(kw_only=True)
class VolumeGrant:
    """"""

    principal: VariableOr[str]

    privileges: VariableOrList[VolumeGrantPrivilege] = field(default_factory=list)

    @classmethod
    def from_dict(cls, value: "VolumeGrantDict") -> "Self":
        return _transform(cls, value)

    def as_dict(self) -> "VolumeGrantDict":
        return _transform_to_json_value(self)  # type:ignore


class VolumeGrantDict(TypedDict, total=False):
    """"""

    principal: VariableOr[str]

    privileges: VariableOrList[VolumeGrantPrivilegeParam]


VolumeGrantParam = VolumeGrantDict | VolumeGrant
